/*==========================================================
 * Program : serv_stuff.c                  Project : smslink
 * Author  : Philippe Andersson.
 * Date    : 05/05/01
 * Version : 0.16b
 * Notice  : (c) Les Ateliers du Heron, 1998 for Scitex Europe, S.A.
 * Comment : Library of functions for the smslink server.
 *
 * Modification History :
 * - 0.01b (19/08/99) : Initial release. Moved the SMS server-
 *   specific functions from stuff.c over here.
 * - 0.02b (21/10/99) : Moved tellsock() over to stuff.c (also
 *   used by sms2mailgw for communication with sendmail).
 * - 0.03b (24/01/00) : The definition for union semun disappeared
 *   from the include files coming with the latest version of
 *   glibc2. Fixed the code for daemons_death() and set_semvalue()
 *   to account for this.
 * - 0.04b (31/03/00) : Adapted gsmdevcpy() to accomodate for
 *   structure change (defmode field).
 * - 0.05b (17/05/00) : Adapted gsmdevcpy() to accomodate for
 *   structure change (capmatrix field).
 * - 0.06b (22/06/00) : Adapted gsmdevcpy() to accomodate for
 *   structure change (capmatrix structure).
 * - 0.07b (26/09/00) : Added a signal handler for SIGCHLD to prevent
 *   zombie formation (contributed by Stanley Hopcroft 
 *   <Stanley.Hopcroft@IPAustralia.Gov.AU>). Modified it to log more
 *   info about the exit status of the child.
 * - 0.08b (03/10/00) : Removed all logging from sig_chld(). The
 *   server died after a few days in production, and it might have
 *   been due to the use of unsafe functions within a signal
 *   handler.
 * - 0.09b (18/02/01) : Included debug level handling. Modified
 *   gsmdevcpy() to accomodate new <flags> field. Improved
 *   debug info in tell_gsm() and get_gsm_answer(). Added fn
 *   cleanup_mdm(). Contributed by Andrew Worsley 
 *   <epaanwo@asac.ericsson.se>.
 * - 0.10b (19/02/01) : Added support for fast modem response,
 *   and added set_cpms() to set the "message store". Contributed 
 *   by Andrew Worsley <epaanwo@asac.ericsson.se>.
 * - 0.11b (05/03/01) : Replaced control of all debug info by
 *   global flags.
 * - 0.12b (30/03/01) : Added hexdump facility for gsm answers,
 *   controlled by global debug level, to ease debugging "fast
 *   modem response" mode.
 * - 0.13b (02/04/01) : Modified gsmdevcpy() to support new
 *   modem_okay field. Passed the struct gsms_def as param. to
 *   get_gsm_answer().
 * - 0.14b (04/04/01) : Added functions for cpl_list handling.
 *   Altered get_gsm_resp(), get_gsm_answer() and get_gsm_fast()
 *   to be passed a cpl_list parameter instead of a simple char*.
 * - 0.15b (03/05/01) : Added getregstat() function.
 * - 0.16b (05/05/01) : Fixed a small bug in the debugging code
 *   in getregstat() function.
 *========================================================*/

#include <unistd.h>
#include <stdio.h>                         /* for fprintf */
#include <stdlib.h>                  /* for errno & stuff */
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <sys/time.h>           /* for the struct timeval */
#include <sys/types.h>
#include <sys/wait.h>                    /* for waitpid() */
#include <sys/ipc.h>
#include <sys/sem.h>                        /* semaphores */
#include <sys/shm.h>                     /* shared memory */
#include <sys/ioctl.h>            /* for the ioctl() call */
#include <dial/modems.h>           /* requires 'libmodem' */

#include "sms_serv.h"

/*========================================================*/
/**********             LOCAL DEFINES              ********/
/*========================================================*/
/* For debugging purposes only - comment out for normal compile */
/* #define INCL_DEBUG_CODE */

/*========================================================*/
/**********           GLOBAL VARIABLES             ********/
/*========================================================*/
extern int debug;                    /* debug level/flags */
extern int use_fast;        /* enable fast modem response */

/*========================================================*/
/**********               FUNCTIONS                ********/
/*========================================================*/
/*========================================================*/
/*################### cpl_list handling ##################*/
/*========================================================*/
void cpl_list_init (cpl_list *list)
{
  list->head = NULL;
  list->tail = NULL;
}                                     /* cpl_list_init () */
/*========================================================*/
int empty_cpl_list (cpl_list list)
{
  return (list.head == NULL);
}                                    /* empty_cpl_list () */
/*========================================================*/
void cpl_list_insert (cpl_list *list, char *c_phrase)
{
  struct cpl_item *element;

  /* alloc memory for new element */
  element = (struct cpl_item *) malloc (sizeof (struct cpl_item));
  if (!element)
    syserr ("sms_serv: can't malloc() for new CPL entry");
  
  /* initialize fields for new element */
  element->catch_phrase = (char *) malloc ((strlen (c_phrase) + 1) * sizeof (char));;
  if (! element->catch_phrase)
    syserr ("sms_serv: can't malloc() for new CPL entry");
  strcpy (element->catch_phrase, c_phrase);
  
  /* chain it in the list */
  if (empty_cpl_list (*list)) {
    list->head = element;
    list->tail = element;
    element->next = NULL;
    element->prev = NULL;
  }
  else {
    element->next = NULL;
    element->prev = list->tail;
    list->tail->next = element;
    list->tail = element;
  }
}                                   /* cpl_list_insert () */
/*========================================================*/
int cpl_match_any (char *s, cpl_list list)
/* Returns TRUE whenever any of the strings stored in list
   is matches a substring of s.
   Returns FALSE otherwise. */
{
  struct cpl_item *cursor;
  int found = FALSE;
  
  if (!empty_cpl_list (list)) {
    cursor = list.head;
    found = (strstr (s, cursor->catch_phrase) != 0);
    while ((cursor->next != NULL) && !found) {
      cursor = cursor->next;
      found = (strstr (s, cursor->catch_phrase) != 0);
    }
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "cpl_match_any(): %s\n",
            (found ? "match found" : "no match found"));
  }
  return (found);
}                                     /* cpl_match_any () */
/*========================================================*/
void dump_cpl_list (cpl_list list)
{
  struct cpl_item *cursor;
  int item = 0;

  if (!empty_cpl_list (list)) {
    cursor = list.head;
    fprintf (stderr, "Item <%d> ", item);
    dump_string (cursor->catch_phrase);
    while (cursor->next != NULL) {
      cursor = cursor->next;
      item++;
      fprintf (stderr, "Item <%d> ", item);
      dump_string (cursor->catch_phrase);
    }
  }
  else {
    fprintf (stderr, "sms_serv: empty 'CPL' list.\n");
  }
}                                     /* dump_cpl_list () */
/*========================================================*/
void free_cpl_list (cpl_list *list)
{
  struct cpl_item *cursor;

  if (!empty_cpl_list (*list)) {
    /* hop to element before last */
    cursor = list->tail->prev;
    /* now go back and clean behind */
    while (cursor != NULL) {
      free (cursor->next->catch_phrase);
      free (cursor->next);
      cursor->next = NULL;
      list->tail = cursor;
      cursor = cursor->prev;
    }                           /* while (cursor != NULL) */
  }                           /* if (!empty_cpl_list (... */
  /* now clean last element and reset header */
  free (list->head->catch_phrase);
  free (list->head);
  list->head = NULL;
  list->tail = NULL;
}                                     /* free_cpl_list () */
/*========================================================*/
/*############ processes and signals handling ############*/
/*========================================================*/
void daemons_death ()
{
#ifdef _SEM_SEMUN_UNDEFINED           /* see <bits/sem.h> */
  union semun {
    int val;				/* value for SETVAL */
    struct semid_ds *buf;		/* buffer for IPC_STAT & IPC_SET */
    unsigned short int *array;		/* array for GETALL & SETALL */
    struct seminfo *__buf;		/* buffer for IPC_INFO */
  } sem_union;
#else  
  union semun sem_union;
#endif                     /* #ifdef _SEM_SEMUN_UNDEFINED */
  extern int shmem_sem, global_sem;
  extern int shmem_id;
  extern void *shared_memory;
  extern acl_list gsm_acl;
  extern int errno;
  
  /* free what's need to be freed */
  free_acl_list (&gsm_acl);
  
  /* release shared memory */
  if (shmdt (shared_memory) == -1)
    syserr ("sms_serv: can't detach from shared memory segment");
  if (shmctl (shmem_id, IPC_RMID, 0) == -1)
    syserr ("sms_serv: can't remove shared memory segment");
  
  /* delete semaphores */
  if (semctl (shmem_sem, 0, IPC_RMID, sem_union) == -1)
    syserr ("sms_serv: can't delete shmem semaphore");
  
  if (semctl (global_sem, 0, IPC_RMID, sem_union) == -1)
    syserr ("sms_serv: can't delete global semaphore");
  
  /* remove checkpoint file */
  if (unlink (CHECKPOINTF) == -1) {
    if (errno != ENOENT) {             /* no such file... */
      syserr ("sms_serv: can't remove checkpoint file");
    }
  }

  /* log it... */
  syslog ((FACILITY | LOG_INFO), "server exiting on SIGTERM.");

  /* closes connection to the syslog daemon */
  closelog ();
  
  /* now exits gracefully */
  exit (0);
}                                     /* daemons_death () */
/*========================================================*/
void sig_chld ()
{
  int stat;
  pid_t pid;
  char cause[20];
  
  if ((pid = waitpid (-1, &stat, 0)) <= 0)
    syserr ("sms_serv: waitpid error in sig_chld()");
  
  /* child's autopsy */
  /* those functions might not be safe to use in a signal handler...
  if (WIFEXITED (stat)) {
    sprintf (cause, "exit code %d", WEXITSTATUS (stat));
  }
  else if (WIFSIGNALED (stat)) {
    sprintf (cause, "signal %d", WTERMSIG (stat));
  }
  else {
    sprintf (cause, "cause unknown");
  }
  */
  
  /* log the event */
  /* those functions might not be safe in a signal handler...
  syslog ((FACILITY | LOG_INFO), "child [%d] just terminated (%s).", pid,
          cause);
  */
  
  /* go on... */
  return;
}                                          /* sig_chld () */
/*========================================================*/
/*################# struct gsms_def handling #############*/
/*========================================================*/
int gsmdevcpy (struct gsms_def *dest, struct gsms_def *src)
{
  if ((src != NULL) && (dest != NULL)) {
    dest->free = src->free;
    dest->owner = src->owner;
    strcpy (dest->device, src->device);
    strcpy (dest->PIN, src->PIN);
    strcpy (dest->PUK, src->PUK);
    strcpy (dest->addr, src->addr);
    strcpy (dest->defsca, src->defsca);
    strcpy (dest->provider, src->provider);
    dest->capmatrix.capflags = src->capmatrix.capflags;
    strcpy (dest->capmatrix.datefmt, src->capmatrix.datefmt);
    strcpy (dest->capmatrix.modem_okay, src->capmatrix.modem_okay);
    strcpy (dest->capmatrix.line_sep, src->capmatrix.line_sep);
    dest->defmode = src->defmode;
    dest->flags = src->flags;
    return (0);
  }
  else {
    return (-1);
  }
}                                         /* gsmdevcpy () */
/*========================================================*/
/*################ Dialogue with GSM module ##############*/
/*========================================================*/
int tell_gsm (int fd, char *command)
{
  int len;
  
  len = strlen (command);
  if (write (fd, command, len) != len) {
    syslog ((FACILITY | LOG_ERR), "error writing to the GSM module.");
    mdmperror ("error writing to the GSM module");
    mdm_unlock (mdmopendevice);
    hangup (fd);
    free (command);
    exit (-1);
  }
  if (debug & DEBUG_TELL) {
    syslog ((FACILITY | LOG_DEBUG), "tell_gsm(%d, \"%s\")", fd, command);
  }
}                                          /* tell_gsm () */
/*========================================================*/
int get_gsm_resp (int fd, char *answer, int size, int resptime,
                 cpl_list cpl)
/*
 * get a response from the GSM modem, will select sleep or fast method
 * dependant on the global variable setting
 */
{
  if (use_fast)
    return (get_gsm_fast (fd, answer, size, resptime, cpl));

  return (get_gsm_sleep (fd, answer, size, resptime));
}                                      /* get_gsm_resp () */
/*========================================================*/
int get_gsm_answer (int fd, char *answer, int size, int resptime,
                    struct gsms_def *gsm)
{
  cpl_list cpl;       /* to store magic terminator string */
  int retval;
  
  cpl_list_init (&cpl);
  cpl_list_insert (&cpl, gsm->capmatrix.modem_okay);
  
  if (use_fast) {
    retval = get_gsm_fast (fd, answer, size, resptime, cpl);
  }
  else {
    retval = get_gsm_sleep (fd, answer, size, resptime);
  }
  
  free_cpl_list (&cpl);
  return (retval);
}                                    /* get_gsm_answer () */
/*========================================================*/
int get_gsm_fast (int fd, char *answer, int size, int resptime,
                 cpl_list cpl)
{
  int nread, retval, previous;
  fd_set inputs;
  struct timeval timeout;
  char *buffer;
  char *answerPtr = answer; /* next available character in the answer area */
      /* last available spot in answer area -> save it for '\0' */
  char *answerEnd = answer + size - 1;
  size_t spaceLeft = size; /* space left in buffer */
  
  /*--------------------------------------Initializations */
  nread = previous = 0;
   
  FD_ZERO (&inputs);
  FD_SET (fd, &inputs);
  
  timeout.tv_sec = 10;
  timeout.tv_usec = 0;
  
  /* non-buffering solution. 
   * reads into the given buffer until terminating string seen or timeout. If
   * too many characters read it falls back to original behavour sleeping and
   * calling FIONREAD and discarding extra characters by reading them into
   * a temporary buffer
   */
  if (debug & DEBUG_HEXDUMP) {
    fprintf (stderr, "Hexdump of current catch phrase list:\n");
    dump_cpl_list (cpl);
  }
  for (;;) {
    if (answerPtr >= answerEnd) /* we have no space left in the buffer */
	break; /* fall through - just read in rest of data and drop it */

    spaceLeft = answerEnd - answerPtr - 1;
    FD_SET (fd, &inputs);
    retval = select (FD_SETSIZE, &inputs, NULL, NULL, &timeout);
    switch (retval) {
      case 0:
        syslog ((FACILITY | LOG_WARNING), "timeout while waiting for GSM answer.");
        return (0);                 /* return zero on error */

      case -1:
        syserr ("sms_serv: call to select() failed");
        return (0);               /* return zero on error */
      
      default:
	break;                           /* got some data */
    }                                  /* switch (retval) */

    ioctl (fd, FIONREAD, &nread);
    if (spaceLeft < nread) /* more data than space -> just fill buffer */
	nread = spaceLeft;
    if ((nread = read (fd, answerPtr, nread)) <= 0) {
      syslog ((FACILITY | LOG_ERR), "read failed: %m");
      return (0);                 /* return zero on error */
    }
    answerPtr[nread] = '\0'; /* terminate the string for strstr() */
    if (debug & DEBUG_GET_GSM_ANSWER) {
      fprintf (stderr, "get_gsm_fast() timeout %d.%d read in %d bytes '%s'\n",
	      timeout.tv_sec, timeout.tv_usec, nread, answerPtr);
      if (debug & DEBUG_HEXDUMP) {
	fprintf (stderr, "Hexdump of receive buffer:\n");
	dump_string (answerPtr);
      }
    }
    answerPtr += nread;
    if (cpl_match_any (answer, cpl)) { /* found end of string */ 
      if (debug & DEBUG_GET_GSM_ANSWER) {
	syslog ((FACILITY | LOG_DEBUG),
	       "get_gsm_answer(%d, \"%s\", %d, %d)",
	       fd, answer, size, resptime);
      }
      return (answerPtr - answer);
    }
  }                                           /* for (;;) */
  /* junk rest of data */
  ioctl (fd, FIONREAD, &nread);
  while (nread != previous) {
      sleep (resptime);
      previous = nread;
      ioctl (fd, FIONREAD, &nread);
  }                          /* while (nread != previous) */
  /* we know what's the data size - alloc space for it */
  buffer = (char *) malloc ((nread + 1) * sizeof (char));
  if (!buffer)
      syserr ("sms_serv: can't allocate buffer space");
  /* now we can finally read this data */
  nread = read (fd, buffer, nread);
  switch (nread) {
    case 0:
      /* EOF */
      syslog ((FACILITY | LOG_WARNING), "got no data from GSM module.");
      break;

    case -1:
      syserr ("sms_serv: error while reading answer from GSM");
      break;

    default:
      buffer[nread] = '\0';
      syslog((FACILITY | LOG_WARNING),
	  "too much data, truncating following %d bytes '%s'", nread, buffer);
      if (debug & DEBUG_GET_GSM_ANSWER) {
	fprintf (stderr, "get_gsm_fast(), truncating following %d bytes '%s'\n",
		nread, buffer);
	if (debug & DEBUG_HEXDUMP) {
	  fprintf (stderr, "Hexdump of receive buffer:\n");
	  dump_string (buffer);
	}
      }
      break;
  }                                     /* switch (nread) */

  free (buffer);
  return (answerPtr - answer);
}                                      /* get_gsm_fast () */
/*========================================================*/
int get_gsm_sleep (int fd, char *answer, int size, int resptime)
{
  int nread, retval, previous;
  fd_set inputs;
  struct timeval timeout;
  char *buffer;
  
  /*--------------------------------------Initializations */
  nread = previous = 0;
  
  FD_ZERO (&inputs);
  FD_SET (fd, &inputs);
  
  timeout.tv_sec = 10;
  timeout.tv_usec = 0;
  
  /* wait for data to arrive on the line */
  retval = select (FD_SETSIZE, &inputs, NULL, NULL, &timeout);
  switch (retval) {
    case 0:
      syslog ((FACILITY | LOG_WARNING), "timeout while waiting for GSM answer.");
      break;
    
    case -1:
      syserr ("sms_serv: call to select() failed");
      break;
      
    default:
      if (FD_ISSET (fd, &inputs)) {
        /* first wait for all data to be ready */
	ioctl (fd, FIONREAD, &nread);
	while (nread != previous) {
	  sleep (resptime);
	  previous = nread;
	  ioctl (fd, FIONREAD, &nread);
	}                    /* while (nread != previous) */
	/* we know what's the data size - alloc space for it */
	buffer = (char *) malloc ((nread + 1) * sizeof (char));
	if (!buffer)
	  syserr ("sms_serv: can't allocate buffer space");
	/* now we can finally read this data */
	nread = read (fd, buffer, nread);
	switch (nread) {
	  case 0:
	    /* EOF */
	    syslog ((FACILITY | LOG_WARNING), "got no data from GSM module.");
	    break;
	    
	  case -1:
	    syserr ("sms_serv: error while reading answer from GSM");
	    break;
	    
	  default:
	    buffer[nread] = '\0';
            if (debug & DEBUG_GET_GSM_ANSWER) {
	      fprintf (stderr, "get_gsm_sleep(), pid<%d> Got : [%s] (%d char)\n",
	              getpid (), buffer, nread);
	      if (debug & DEBUG_HEXDUMP) {
		fprintf (stderr, "Hexdump of receive buffer:\n");
		dump_string (buffer);
	      }
            }
	    /* here we could pre-process it (remove Ctrl-Z) */
	    /* copy it over to out string param. */
	    if (nread > size) {
	      syslog ((FACILITY | LOG_WARNING), "too much data, truncation will occur");
	      strncpy (answer, buffer, size);
	      answer[size] = '\0';
	    }
	    else {
	      strcpy (answer, buffer);
	      if (debug & DEBUG_GET_GSM_ANSWER) {
		syslog ((FACILITY | LOG_DEBUG),
		       "get_gsm_answer(%d, \"%s\", %d, %d)",
		       fd, answer, size, resptime);
	      }
	    }
	    break;
	}                               /* switch (nread) */
      }                                /* if (FD_ISSET... */
      free (buffer);
      break;
  }                                    /* switch (retval) */
  return (nread);
}                                     /* get_gsm_sleep () */
/*========================================================*/
int set_cpms (struct gsms_def *gsm, int fd)
/*
 * force preferred message storage area to be ME - which is only area
 * accessible via AT+CMGL on the GM-12.
 */
{
  char* scratch = (char *) malloc ((BIGBUFF + 1) * sizeof (char));

  if (!scratch)
    syserr ("sms_daemon: can't allocate scratch space");
  memset (scratch, 0, (BIGBUFF + 1));
  
  /* set "message store for reading" to where incomming messages are stored ME */
  sprintf (scratch, "AT+CPMS=\"ME\",\"ME\"\r");
  tell_gsm (fd, scratch);
  memset (scratch, 0, (BIGBUFF + 1));
  if (get_gsm_answer (fd, scratch, BIGBUFF, 1, gsm)) {
    /* check for "OK" */
    if (strstr (scratch, "OK") == NULL) {
      free (scratch);
      syslog ((FACILITY | LOG_ERR), "error after sending AT+CNMI command.");
      mdmperror ("sms_serv: error after sending AT+CNMI command");
      return (-1);
    }
  }
  else {
    free (scratch);
    syslog ((FACILITY | LOG_ERR), "GSM module not responding.");
    mdmperror ("sms_serv: GSM not responding");
    return (-1);
  }

  free (scratch);
  return (0);
}                                          /* set_cpms () */
/*========================================================*/
int getregstat (char *answer)
/*
 * Parses the +CREG: answer and returns the <stat> value
 * (= registration status).
 * Returns -1 on non-parseable input.
 */
{
  char *p1;
  char *p2;
  char *p3;
  int stat;
  int mode;
  
  /* find beginning of "+CREG:" token */
  if ((p1 = strstr (answer, "+CREG:")) == 0) {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "getregstat(): can't find +CREG: header in [%s]\n",
             answer);
      if (debug & DEBUG_HEXDUMP) {
	fprintf (stderr, "Hexdump of +CREG answer:\n");
	dump_string (answer);
      }
    }
    return (-1);
  }
  
  /* Hop to first char after the ':' (should be a WS) */
  p1 += 6;
  
  /* Locate the (first ?) comma */
  if ((p2 = strchr (p1, ',')) == 0) {
    /* there should be at least 2 fields */
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "getregstat(): can't locate start of 2nd field in +CREG: string [%s]\n",
             answer);
      if (debug & DEBUG_HEXDUMP) {
	fprintf (stderr, "Hexdump of +CREG answer:\n");
	dump_string (answer);
      }
    }
    return (-1);
  }
  
  /* Terminate the <mode> field */
  if (strlen (p2) <= 1) {
    /* nothing after the comma */
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "getregstat(): 2nd field looks empty in +CREG: string [%s]\n",
             answer);
      if (debug & DEBUG_HEXDUMP) {
	fprintf (stderr, "Hexdump of +CREG answer:\n");
	dump_string (answer);
      }
    }
    return (-1);
  }
  p2[0] = '\0';
  
  /* p1 is now <mode> field (ignored) */
  mode = atoi (p1);
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "getregstat(): found <mode> field: [%d] (ignored)\n",
           mode);
  }
  
  /* Hop to the first char. of the <stat> field */
  p2++;
  
  /* skip any whitespace */
  while ((p2[0] == ' ') || (p2[0] == '\t')) {
    p2++;
  }
  
  /* locate the end of the <stat> field */
  /* any additional field or non-digit char. is ignored */
  p3 = p2;
  while (isdigit (p3[0])) {
    p3++;
  }

  /* Terminate the <stat> field */
  p3[0] = '\0';
  
  /* p2 is now <stat> field -- convert it */
  stat = atoi (p2);
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "getregstat(): found <stat> field: [%d] (returned)\n",
           stat);
  }
  
  return (stat);
}                                        /* getregstat () */
/*========================================================*/
/*####################### Semaphores #####################*/
/*========================================================*/
int set_semvalue (int sem_id, int value)
/* returns 0 on success, -1 on failure */
{
#ifdef _SEM_SEMUN_UNDEFINED		/* see <bits/sem.h> */
  union semun {
    int val;				/* value for SETVAL */
    struct semid_ds *buf;		/* buffer for IPC_STAT & IPC_SET */
    unsigned short int *array;		/* array for GETALL & SETALL */
    struct seminfo *__buf;		/* buffer for IPC_INFO */
  } sem_union;
#else  
  union semun sem_union;
#endif                     /* #ifdef _SEM_SEMUN_UNDEFINED */
  
  sem_union.val = value;
  return (semctl (sem_id, 0, SETVAL, sem_union));
}                                      /* set_semvalue () */
/*========================================================*/
int sem_wait (int sem_id)
/* waits for the semaphore to be decreased */
{
  struct sembuf sem_b;
  
  sem_b.sem_num = 0;                   /* sem array index */
  sem_b.sem_op = -1;             /* value to inc. sem. by */
  sem_b.sem_flg = SEM_UNDO;
  
  /* --- can sleep here --- */
  return (semop (sem_id, &sem_b, 1));
}                                      /* int sem_wait () */
/*========================================================*/
int sem_decreq (int sem_id)
/* attempts to decrease the semaphore - return error if fails */
{
  struct sembuf sem_b;
  
  sem_b.sem_num = 0;                   /* sem array index */
  sem_b.sem_op = -1;             /* value to inc. sem. by */
  sem_b.sem_flg = (IPC_NOWAIT | SEM_UNDO);
  
  return (semop (sem_id, &sem_b, 1));
}                                    /* int sem_decreq () */
/*========================================================*/
int sem_signal (int sem_id)
/* increases the semaphore - always succeeds */
{
  struct sembuf sem_b;
  
  sem_b.sem_num = 0;                   /* sem array index */
  sem_b.sem_op = 1;              /* value to inc. sem. by */
  sem_b.sem_flg = SEM_UNDO;
  
  return (semop (sem_id, &sem_b, 1));
}                                    /* int sem_signal () */
/*========================================================*/
/*######################### Misc #########################*/
/*========================================================*/
void cleanup_mdm (const char *msg, int fd)
/*
 * report error and close the given modem line if the file descriptor is >= 0
 * and the global modem device. Can be used with a null message to close down
 * the connection properly when there is no error.
 */
{
    if (msg) {
      /* needed straight away before error code is overwritten ?? */
      mdmperror ((char *) msg);
      syslog ((FACILITY | LOG_ERR), msg);
    }
    mdm_unlock (mdmopendevice);
    if (fd >= 0) {
      hangup (fd);
    }
}                                       /* cleanup_mdm () */
/*==========================================================
 * EOF : serv_stuff.c
 *===================*/

